#
# Copyright (c) Daniel Sheffield 2023
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from typing import Iterable
import os
from urllib.parse import urlencode
from bottle import (
    route,
    request,
    run,
    response,
    FormsDict,
    redirect,
    template,
    static_file,
    TEMPLATE_PATH,
)
from matplotlib.axes import Axes
from psycopg import connect
from psycopg.sql import SQL, Literal
import seaborn as sns
from multiprocessing import Lock

from ..data.filter import(
    get_filter,
)

from ..data.util import(
    get_where_include_exclude
)
from .trend import trend_internal
from . import PARAMS
import matplotlib
matplotlib.use('agg')

CACHE = dict()

host = f"host={os.getenv('HOST')}"
db = f"dbname={os.getenv('DB', 'grocery')}"
user = f"user={os.getenv('USER', 'das')}"
password = f"password={os.getenv('PASSWORD','')}"
if not password.split('=',1)[1]:
    password = ''
conn = connect(f"{host} {db} {user} {password}")
sns.set_theme(style='darkgrid', palette='pastel')

def get_product_rollup_statement(filters, having=None):
    where = [ get_where_include_exclude(
        k[0], "name", list(include), list(exclude)
    ) for k, (include, exclude) in filters.items() ]
    return SQL('\n').join([
        SQL("""
SELECT
  count(DISTINCT p.id) AS "Products",
  count(DISTINCT c.id) AS "Categories",
  count(DISTINCT g.id) AS "Groups",
  p.name AS "Product",
  c.name AS "Category",
  g.name AS "Group"
FROM products p
JOIN categories c ON p.category_id = c.id
JOIN groups g ON c.group_id = g.id
"""),
        SQL("""
WHERE {where}
""").format(where=SQL("\nAND").join(where)) if where else SQL(''),
        SQL("""
GROUP BY ROLLUP (g.name, c.name, p.name)
"""),
        SQL("""
HAVING {having}
ORDER BY "Product", "Category", "Group"
""").format(having=having) if having else SQL('')
    ])


def get_query_param(include, exclude):
    return '!'.join([
        '|'.join(sorted(include)),
        *([ '|'.join(sorted(exclude)) ] if exclude else [ ]),
    ])

def normalize_query(query: FormsDict, allow: Iterable[str] = None) -> str:
    param = get_filter(query, allow=allow)
    return urlencode([
        (k, get_query_param(*param[k])) for k in sorted(param) if param[k]
    ])



@route('/grocery/static/<filename:path>')
def send_static(filename):
    return static_file(filename, root='app/rest/static')


global LOCK
LOCK = Lock()

@route('/grocery/trend')
def trend():
    _, _, path, *_ = request.urlparts
    normalized = normalize_query(request.query, allow=PARAMS)

    if request.query_string != normalized:
        return redirect(f'{path}?{normalized}')
    
    loading = template("loading", progress=[])
    if request.query_string in CACHE:
        if LOCK.acquire(block=False):
            try:
                return CACHE[request.query_string]["state"]
            finally:
                try:
                    CACHE[request.query_string]["state"] = next(CACHE[request.query_string]["iter"])
                except StopIteration:
                    del CACHE[request.query_string]
                finally:
                    LOCK.release()
        else:
            return CACHE[request.query_string]["state"]

    if LOCK.acquire(block=False):
        if request.query_string in CACHE:
            LOCK.release()
            return CACHE[request.query_string]["state"]
        try:
            CACHE[request.query_string] = {
                "iter": trend_internal(conn, path, request.query),
                "state": loading,
            }
        finally:
            LOCK.release()
    return loading


@route('/grocery/groups')
def groups():
    filters = get_filter(request.query, allow=('group', 'category', 'product'))
    form = template('form-nav', action='groups', method='get', params=[
        {'name': k, 'value': request.params[k]} for k in request.params if k in PARAMS
    ])
    try:
        with conn.cursor() as cur:
            inner = get_product_rollup_statement(
                filters,
                having=SQL("c.name IS NULL")
            )
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({q}, false, false, ''::text)
""").format(q=Literal(SQL("""SELECT
    "Products",
    "Categories",
    COALESCE("Group", "Groups"||'') "Group"
FROM (
{inner}
) q""").format(inner=inner).as_string(cur)))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template("query-to-xml", title="Groups", xml=xml, form=form)

@route('/grocery/categories')
def categories():
    filters = get_filter(request.query, allow=('group', 'category', 'product'))
    form = template('form-nav', action='categories', method='get', params=[
        {'name': k, 'value': request.params[k]} for k in request.params if k in PARAMS
    ])
    try:
        with conn.cursor() as cur:
            inner = get_product_rollup_statement(
                filters,
                having=SQL("p.name IS NULL AND (c.name IS NOT NULL OR g.name IS NULL)")
            )
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({q}, false, false, ''::text)
""").format(q=Literal(SQL("""SELECT
    "Products",
    COALESCE("Category", "Categories"||'') "Category",
    COALESCE("Group", "Groups"||'') "Group"
FROM (
{inner}
) q""").format(inner=inner).as_string(cur)))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template("query-to-xml", title="Categories", xml=xml, form=form)

@route('/grocery/products')
def products():
    filters = get_filter(request.query, allow=('group', 'category', 'product'))
    form = template('form-nav', action='products', method='get', params=[
        {'name': k, 'value': request.params[k]} for k in request.params if k in PARAMS
    ])
    try:
        with conn.cursor() as cur:
            inner = get_product_rollup_statement(
                filters,
                having=SQL("p.name IS NOT NULL OR g.name IS NULL")
            )
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({q}, false, false, ''::text)
""").format(q=Literal(SQL("""SELECT
    --"Transactions",
    COALESCE("Product", "Products"||'') "Product",
    COALESCE("Category", "Categories"||'') "Category",
    COALESCE("Group", "Groups"||'') "Group"
FROM (
{inner}
) q""").format(inner=inner).as_string(cur)))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template("query-to-xml", title="Products", xml=xml, form=form)

@route('/grocery/tags')
def tags():
    form = template('form-nav', action='tags', method='get', params=[
        {'name': k, 'value': request.params[k]} for k in request.params if k in PARAMS
    ])
    try:
        with conn.cursor() as cur:
            inner = SQL('\n').join([SQL("""
SELECT * FROM (SELECT count(DISTINCT txn.id) AS "Uses", tg.name AS "Name"
FROM tags tg
JOIN tags_map tm ON tg.id = tm.tag_id
JOIN transactions txn ON txn.id = tm.transaction_id
GROUP BY tg.name
ORDER BY 1 DESC, 2) q
UNION ALL
SELECT count(DISTINCT txn.id) AS "Uses", count(DISTINCT tg.name)||'' AS "Name"
FROM tags tg
JOIN tags_map tm ON tg.id = tm.tag_id
JOIN transactions txn ON txn.id = tm.transaction_id
""")]).as_string(cur)
            xml = cur.execute(SQL("""
SELECT query_to_xml_and_xmlschema({inner}, false, false, ''::text)
""").format(inner=Literal(inner))).fetchone()[0]
    finally:
        conn.commit()
    response.content_type = 'application/xhtml+xml; charset=utf-8'
    return template("query-to-xml", title="Tags", xml=xml, form=form)

run(host='0.0.0.0', port=6772, server='gunicorn')
