#
# Copyright (c) Daniel Sheffield 2023
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from bottle import (
    DictProperty,
    HTTPError,
    template,
    request,
)
from io import StringIO
import matplotlib.pyplot as plt
import seaborn as sns
from psycopg import Connection
from psycopg.connection import TupleRow
from time import time
from . import ALL_UNITS
from ..data.QueryManager import (
    display_mapper,
    QueryManager,
)
from ..data.filter import (
    get_filter,
)
from ..activities.Plot import (
    get_data,
)
from .form import(
    get_form,
)
from . import PARAMS

def abort(code, text):
    return HTTPError(code, text)

def trend_internal(conn: Connection[TupleRow], path: str, query: DictProperty):
    progress = []
    try:
        with conn.cursor() as cur:
            query_manager = QueryManager(cur, display_mapper)
            fields = { k: query[k] or None for k in query.keys() if k in PARAMS }
            unit = fields['unit'] = fields['unit'] or 'kg' if 'unit' in fields else 'kg'
            if unit and unit not in ALL_UNITS:
                raise abort(400, f"Unsupported unit {unit}")

            progress.append({ "name": "Loading data", "status": ""})
            yield template("loading", progress=progress)
            data = get_data(query_manager, **fields)

            progress[-1]["status"] = "done"
            yield template("loading", progress=progress)

            if data.empty:
                raise abort(404, f"No data for {fields}")
            
            progress.append({ "name": "Loading chart", "status": ""})
            yield template("loading", progress=progress)
            
            pivot = data.pivot_table(index=['ts_raw',], columns=['product',], values=['$/unit'], aggfunc='mean')
            pivot.columns = pivot.columns.droplevel()
            
            sns.set(style="darkgrid", palette='pastel', context="talk")
            plt.style.use("dark_background")
            plt.rcParams.update({"grid.linewidth":0.1, "grid.alpha":0.5})
            plt.figure(figsize=[16, 9], layout="tight")
            xlabel='Time'
            ylabel=f'$ / {unit}'
            if pivot.columns.size > 50:
                ax = sns.scatterplot(data=pivot, markers=True)
            else:
                ax = sns.lineplot(data=pivot, markers=True)
                legend = plt.figlegend(
                    loc='upper center', ncol=6,
                    title_fontsize="14", fontsize="12", labelcolor='#a0a0a0',
                    framealpha=0.5
                )
                legend.set_title(title="Products")
            ax.legend().set_visible(False)

            ax.set_xlabel(xlabel, color='#a0a0a0', fontsize="14")
            ax.set_ylabel(ylabel, color='#a0a0a0', fontsize="14")
            ax.axes.tick_params(colors='#a0a0a0', labelsize="12", which='both')
            for _, spine in ax.spines.items():
                spine.set_color('#a0a0a0')
            
            progress[-1]["status"] = "done"
            yield template("loading", progress=progress)
            
            progress.append({ "name": "Rendering chart", "status": ""})
            yield template("loading", progress=progress)

            f = StringIO()
            plt.savefig(f, format='svg')
            _filter = get_filter(request.query, allow=PARAMS)
            form = get_form(path.split('/')[-1], 'get', _filter, data)
            
            progress[-1]["status"] = "done"
            yield template("loading", progress=progress)
            
            resp = lambda: template("trend", form=form, svg=f.getvalue())

    except HTTPError as e:
        resp = lambda exception=e: exception

    finally:
        conn.commit()
     
    yield from iter(resp, lambda started=time(): time() - started > 600)
